import pysam
import pandas as pd
import numpy as np
from collections import defaultdict
from math import ceil
from Bio import SeqIO
import sys
import logging
from ancient_helper_kit.util.general import create_stdout_logger

def get_buckets(arr):

    # to make this comparable, compute the percentage of reads
    # in each of 100 buckets of the genome, so you can say 10% of the reads are in the first 2% of the genome
    step_size = int(ceil(len(arr) / 100))
    res = np.zeros(100)

    generator = (slice(i, min(i + step_size, len(arr))) for i in range(0, len(arr), step_size))
    for i, sl in enumerate(generator):
        res[i] = arr[sl].sum() / arr.sum()
    return ','.join(res.astype(str))


def standardize(arr):
    return (arr-arr.mean())/arr.std()


def sample_coverage_and_call_agg_functions(dic, row, agg_funcs, no_samples=50, sample_ratio=0.66):
    """
    Given a row of the dataframe, that represents a virus-accession, this function looks up the found
    alignment from the bam file via the passed dictionary.
    It then creates a vector representing the coverage found for the respective full virus genome.
    From this vector samples are taken multiple times. The provided functions are then applied on these
    samples. These functions should compute aggregations of the coverage samples, like the mean/median/std.
    @param dic: dictionary with key: NC accession value: pysam alignment
    @param row: a row from the dataframe, containing information about the length and nc accesion of the virus
    @param agg_funcs: function that takes as input a 2D numpy array
    @param no_samples: the number of samples that should be drawn
    @param sample_ratio: the number of locations across the genome that should be assembled, specified as ratio of
    the full genome length
    @return: list with output of the passed functions
    """
    if len(dic[row.name]) == 0:  # not a single alignment found
        return [0 for f in agg_funcs]

    virus_length = row.length
    # create an array to compute the coverage at each position
    coverage_arr = np.zeros(virus_length)
    for m in dic[row.name]:
        coverage_arr[m.get_reference_positions()] += 1

    standardized_cov_arr = standardize(coverage_arr)

    # considering everything above of x times the standard deviation as an outlier
    cleaned5 = standardized_cov_arr[standardized_cov_arr < 5]
    cleaned3 = standardized_cov_arr[standardized_cov_arr < 3]
    cleaned7 = standardized_cov_arr[standardized_cov_arr < 7]

    res = [f(cleaned3) for f in agg_funcs]
    res += [f(cleaned5) for f in agg_funcs]
    res += [f(cleaned7) for f in agg_funcs]
    res.append(get_buckets(coverage_arr))
    res.append((coverage_arr >= 1).sum()/virus_length)
    return res


if __name__ == '__main__':

    with open(snakemake.log.stderr, 'w') as fp_stderr:
        with open(snakemake.log.stdout, 'w') as fp_stdout:
            sys.stderr = fp_stderr
            sys.stdout = fp_stdout
            my_logger = create_stdout_logger()
            my_logger.info('start to read in alignment')
            alignment_lookup = defaultdict(list)
            for alignment in pysam.AlignmentFile(snakemake.input.bam, "rb"):
                alignment_lookup[alignment.reference_name].append(alignment)
            my_logger.info('done')
            if len(alignment_lookup) > 0:
                fasta_sequences = SeqIO.parse(open(snakemake.input.virus_fasta),'fasta')
                temp_list = [[fasta.id, fasta.description, len(fasta.seq)] for fasta in fasta_sequences]

                df1 = pd.DataFrame(temp_list, columns=['nc_accession', 'description', 'length'])
                df2 = pd.DataFrame([[k, len(v)] for (k,v) in alignment_lookup.items()], columns=['nc_accession', 'no_mappings'])
                df = df1.merge(df2, left_on="nc_accession", right_on="nc_accession")
                df.set_index('nc_accession', inplace=True)
                my_logger.info('merging done')

                agg_funcs = [('std', np.std),
                             ('mean', np.mean),
                             ('95_perc', lambda x: np.quantile(x,.95)),
                             ]
                new_cols = [f"{name}_{n}" for n in ["3", "5","7"] for name, _ in agg_funcs ]
                new_cols.append('read_coverage_buckets')
                new_cols.append('breadth')
                df[new_cols] = df.apply(lambda row: sample_coverage_and_call_agg_functions(
                    dic=alignment_lookup, row=row, agg_funcs=[f for _,f in agg_funcs]),
                                              axis=1, result_type='expand')

                my_logger.info('all stats computation done')

                df['map_quality'] = df.apply(lambda x: [_.mapping_quality for _ in alignment_lookup[x.name]], axis=1)
                df['map_quality_mean'] = df.map_quality.apply(lambda x: np.array(x).mean())
                df['map_quality_median'] = df.map_quality.apply(lambda x: np.quantile(
                    np.array(x).astype(int), .5))

                df['ratio_map_to_len'] = df.no_mappings/df.length

                my_logger.info('writing result to csv')
                df.to_csv(snakemake.output.csv)
            else: # dirty hack to get around an empty alignment
                cols = ['nc_accession', 'description', 'length', 'no_mappings', 'std_3',
                        'mean_3', '95_perc_3', 'std_5', 'mean_5', '95_perc_5', 'std_7',
                        'mean_7', '95_perc_7', 'read_coverage_buckets', 'breadth',
                        'map_quality', 'map_quality_mean', 'map_quality_median',
                        'ratio_map_to_len']
                t = pd.DataFrame([['nc_empty', 'empty description'] +
                              [0] * (len(cols) - 2)], columns=cols).set_index('nc_accession')
                t.to_csv(snakemake.output.csv)


import pytest
import ancient_helper_kit.kraken2.create_viral_dna_fasta as cvdf
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq


class TestGenerator:
    seq = Seq("GTTGGCCTACTGGTCGGTCAGTAAAAACTACATCAAGATCCGGTAATT")
    id = "NC_055757.1|kraken:taxid|2250214"
    desc = f"{id} Escherichia virus KFS-EC, complete genome"
    seq_rec = SeqRecord(seq=seq, id=id, name=id, description=desc)
    seq_dict = {"NC_055757.1": seq_rec}
    tax_lookup = {"NC_055757.1": "2250214"}

    def test_same_seq_returned(self):
        rec = next(cvdf.my_generator(seq_dict=self.seq_dict,
                                     tax_lookup=self.tax_lookup))
        assert rec.seq == self.seq

    def test_id(self):
        rec = next(cvdf.my_generator(seq_dict=self.seq_dict,
                                     tax_lookup=self.tax_lookup))
        assert rec.id == self.id

    def test_description(self):
        rec = next(cvdf.my_generator(seq_dict=self.seq_dict,
                                     tax_lookup=self.tax_lookup))
        assert rec.description == self.desc

    def test_empty_description(self):
        self.seq_rec.description = self.id
        rec = next(cvdf.my_generator(seq_dict=self.seq_dict,
                                     tax_lookup=self.tax_lookup))
        assert rec.description == self.id




"""
Collection of general helper functions.
"""

from subprocess import run, CalledProcessError
import sys
import logging


def download_via_http_link(http_link, out_fn=None):
    """
    Given a http link, this function downloads content from the link
    calling wget in a separate process and optionally downloads to a specified location.
    If no location is specified, it will just download to the current directory.
    @param http_link: link to download from
    @param out_fn: output file name
    @return:
    """
    params = ['wget', f"{http_link}", '-O', f"{out_fn}"] if out_fn else ['wget', f"{http_link}"]
    try:
        run(params,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True)
    except CalledProcessError as ex:
        print(f"filter-fasta.py  returned with exit code \
                            {ex.returncode}", file=sys.stderr)
        exit(ex.returncode)


def is_gz_file(filepath):
    """
    Hacky test for the magic number in the 2 first bytes of the file to determine if its gzipped.
    @param filepath:
    @return:
    """
    with open(filepath, 'rb') as test_f:
        return test_f.read(2) == b'\x1f\x8b'


def create_stdout_handler(level=logging.DEBUG):
    """
    Creates a logger handler that writes to stdout.
    Just nicer than just printing to std out, since it includes time stamps
    @return:
    """
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    return handler


def create_stdout_logger(level=logging.DEBUG):
    """
    Creates a logger that writes to stdout.
    Just nicer than just printing to std out, since it includes time stamps
    @return:
    """
    my_logger = logging.getLogger()
    my_logger.setLevel(logging.DEBUG)
    my_logger.addHandler(create_stdout_handler(level))
    return my_logger


def compress(input_fn, output_fn=None):
    """
    Compresses the input file using gzip and remove the original file afterwards.
    If the optional output filename is specified the input file is kept and the compressed
    file is written to a newly created file with the given output filename.
    :param input_fn: File to be compressed
    :param output_fn: file name of the output file to be created
    :return:
    """
    try:
        if output_fn is None:
            run(['gzip', input_fn],
                stdout=sys.stdout,
                stderr=sys.stderr,
                check=True)
        else:
            run(['gzip -c', input_fn],
                stdout=output_fn,
                stderr=sys.stderr,
                check=True)

    except CalledProcessError as ex:
        print(f"gzip returned with exit code \
                            {ex.returncode}", file=sys.stderr)
        exit(ex.returncode)



import pandas as pd
from taxonomy import Taxonomy, TaxonomyError
import sys
from collections import defaultdict
from ancient_helper_kit.util.general import create_stdout_logger

from multiprocessing import Process, Queue


def get_c_q(read_descr,label, tax, taxonomy_err_dic):
    c,q = 0,0
    kmer_info = read_descr.split(' ')
    # test if the tax-id of the read is in my taxonomy
    if tax.node(label) is not None:
        for p in kmer_info:
            kmer_tax, no_kmers = p.split(':')
            if kmer_tax != 'A':
                if kmer_tax != '0':
                    try:
                        if tax.lca(kmer_tax, label).id == label:
                            c += int(no_kmers)
                    except TaxonomyError:
                        taxonomy_err_dic[kmer_tax] += 1
                    q += int(no_kmers)
                elif kmer_tax == '0':
                    q += int(no_kmers)
    else:
        taxonomy_err_dic[label] += len(kmer_info)
    return c,q


def compute_conf_score(input_fn, is_paired, output_fn, tax, queue):
    tax_err_dic = defaultdict(int)
    tl = []
    for line in open(input_fn):
        _, read, name, lengths, kmer_desc = line.strip().split("\t")
        label = name.strip(')').split('taxid ')[-1].strip()
        if is_paired:
            c, q = 0, 0
            for s in map(str.strip, kmer_desc.split('|:|')):
                if len(s) > 0:
                    ct, qt = get_c_q(read_descr=s,
                                     label=label, tax=tax,
                                     taxonomy_err_dic=tax_err_dic)
                    c += ct
                    q += qt
        else:
            c, q = get_c_q(read_descr=kmer_desc,
                           label=label,
                           taxonomy_err_dic=tax_err_dic)
        conf_score = c / q if q > 0 else 0
        tl.append([read, name, conf_score])

    pd.DataFrame(tl, columns=['read_name', 'tax_id', 'conf_score']).to_csv(output_fn,index=False)
    queue.put((is_paired,tax_err_dic))


def main(fn_tax_nodes, fn_tax_names, k2_classified_out_single, csv_single,
         k2_classified_out_paired, csv_paired, logger=None):
    tax = Taxonomy.from_ncbi(nodes_path=fn_tax_nodes,
                             names_path=fn_tax_names)

    if logger is not None:
        logger.info("loaded taxonomy")

    queue = Queue()
    p_single = Process(target=compute_conf_score, args=(k2_classified_out_single,
                                                        False, csv_single, tax, queue))
    p_single.start()

    p_paired = Process(target=compute_conf_score, args=(k2_classified_out_paired,
                                                        True, csv_paired, tax, queue))
    p_paired.start()

    p_paired.join()
    p_single.join()
    if logger is not None:
        logger.info("taxonomy errors of each run:")
        for i in range(2):
            is_paired, d = queue.get()
            logger.info(f"{'paired' if is_paired else 'single'}\n {d}")








import os
import tempfile
import pandas as pd
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import gzip
import argparse
from ancient_helper_kit.util.general import download_via_http_link, is_gz_file, create_stdout_logger
import sys
from multiprocessing import Process, Queue
from typing import Set, Dict

taxonomy_acc2taxId = "https://ftp.ncbi.nih.gov/pub/taxonomy/accession2taxid/nucl_gb.accession2taxid.gz"
genbank_and_refseq_nucleotide_sequences = "http://ftp.ncbi.nlm.nih.gov/genomes/Viruses/AllNucleotide/AllNucleotide.fa"
genbank_and_refseq_nucleotide_meta_data = "https://ftp.ncbi.nlm.nih.gov/genomes/Viruses/AllNuclMetadata/AllNuclMetadata.csv"

meta_data_dtypes = {"#Accession": str,
                     "SRA_Accession": str,
                     "Submitters": str,
                     "Release_Date": str,
                     "Species": str,
                     "Genus": str,
                     "Family": str,
                     "Molecule_type": str,
                     "Length": int,
                     "Sequence_Type": str,
                     "Nuc_Completeness": str,
                    "Genotype": str,
                    "Segment": str,
                    "Publications": float,
                    "Geo_Location": str,
                    "USA": str,
                    "Host": str,
                    "Isolation_Source": str,
                    "Collection_Date": str,
                    "BioSample": str,
                    "GenBank_Title": str,
                    "phage": bool}


def handle_meta_data(in_fn):
    """
    Filter the meta data file down to DNA viruses and try to exclude phages.
    @param in_fn: csv file from NCBI giving meta info about the viral genome sequences from refseq/genbank
    @return: dataframe containing the csv file filtered down
    """
    df_meta = pd.read_csv(in_fn, dtype=meta_data_dtypes)

    def is_phage(row):
        return (("phage" in str(row.Species).lower())
                or "bacteria" in str(row.Host).lower()
                or "bacterium" in str(row.Host).lower()
                or "phage" in str(row.GenBank_Title).lower()
                or "bacterium" in str(row.GenBank_Title).lower())

    df_meta['is_phage'] = df_meta.apply(is_phage, axis=1)
    df_meta['is_dna_virus'] = df_meta.Molecule_type.apply(lambda x: "dna" in str(x).lower())

    df_complete = df_meta[(df_meta.Nuc_Completeness == 'complete')
                          & (~df_meta.is_phage)
                          & df_meta.is_dna_virus]
    return df_complete


def filter_nuc_sequences(nuc_fn: str, filter_set: Set[str], queue=None) -> Dict[str, SeqRecord]:
    """
    Filters out FASTA records from a FASTA file whose Ids are not in the given set and
    returns the records in a dict[str, SeqRecord] with the fasta id as key.
    :param nuc_fn: file name of the fasta file
    :param filter_set: set with allowed ids (whitelist)
    :param queue: necessary if called within a separate process to return the dict
    :return: dict[str, SeqRecord] with the fasta id as key
    """
    ncbi_dict = dict()
    for record in SeqIO.parse(open(nuc_fn), 'fasta'):
        accession = record.id.split('|')[0]
        if accession in filter_set:
            ncbi_dict[accession] = record
    queue.put(('seq_dic', ncbi_dict)) if queue else None
    return ncbi_dict


def read_in_acc_to_taxid(in_fn, filter_set=None, queue=None):
    """
    Iterates over the content of the accession2taxid file that links
    accession ids to tax ids and returns a dictionary with accession:[taxId]
    If filter_set is given only accessions that are in the set are returned.
    @param in_fn:
    @param filter_set:
    @param queue: needs to be specified if run in a separate process to be able to return sth
    @return:
    """
    rd = dict()
    f = gzip.open(in_fn, 'rt') if is_gz_file(in_fn) else open(in_fn)
    with f as fp:
        # accession accession.version   taxid   gi
        fp.readline()  # skip header
        for line in fp:
            _, accession, tax_id, _ = line.strip().split('\t')
            if filter_set:
                if accession in filter_set:
                    assert accession not in rd, f"accession {accession} occurred twice in input file"
                    rd[accession] = tax_id
            else:
                assert accession not in rd, f"accession {accession} occurred twice in input file"
                rd[accession] = tax_id
    queue.put(('tax_id', rd)) if queue is not None else None
    return rd


def my_generator(seq_dict, tax_lookup, meta_df=None):
    """
    Generator taking a dictionary of fasta sequences and a dictionary with accession number
    as key and tax id as value.
    The generator iterates over the dictionary content, if the accession number of the sequence
    is in the tax_lookup dictionary it will add a kraken-specific signature to the fasta description
    that includes the tax-id and yield that modified fasta entry.
    @param seq_dict: dictionary containing fasta sequences {accession: SeqRecord}
    @param tax_lookup: dictionary containing tax-ids entries {accession: tax-id}
    @param meta_df: dataframe with meta information of the sequence
    @return: yields fasta-records with a kraken-specific description
    """
    for k, record in seq_dict.items():
        if k in tax_lookup:
            tax_id = tax_lookup[k]
            new_name = f"{k}|kraken:taxid|{tax_id}"
            if meta_df is not None:
                df = meta_df[meta_df["#Accession"] == k]
                assert len(df) == 1, f"meta_df contains more than one entry or none for accession: {record.id}"
                new_description = f"{new_name} {df.GenBank_Title.values[0]}"
            else:
                # extracting the description only without the id
                t = record.description.split(" ", maxsplit=1)
                new_description = new_name if len(t) == 1 else f"{new_name} {t[-1]}"
            yield SeqRecord(record.seq, id=new_name, description=new_description)


def parse_meta_data(args, my_logger):
    """
    Either downloads or gets metadata about the corresponding fasta file.
    Then it filters down to my chosen criteria.
    :param args: args from the argparser
    :param my_logger: reference to a logger
    :return: filtered data frame of the meta data
    """
    if args.meta:
        my_logger.info('meta data provided, reading and filtering it next ...')
        meta_df = handle_meta_data(args.meta)
    else:
        my_logger.info('meta data not provided, downloading it next ...')
        fp = tempfile.NamedTemporaryFile(delete=False)
        download_via_http_link(genbank_and_refseq_nucleotide_meta_data, out_fn=fp.name)
        my_logger.info('downloading done, next filtering....')
        meta_df = handle_meta_data(fp.name)
        fp.close()
        os.remove(fp.name)
    my_logger.info('finished handling of meta data')
    if args.out_fn_meta:
        my_logger.info(f"writing out meta data to: {args.out_fn_meta}")
        meta_df.to_csv(args.out_fn_meta, index=False)
    return meta_df


def filter_tax_id_mapping(args, my_logger, my_queue, sub_set, temp_file_list):
    # tax id mapping
    if args.acc2tax:
        p1 = Process(target=read_in_acc_to_taxid, args=(args.acc2tax,), kwargs={"filter_set": sub_set,
                                                                                "queue": my_queue})
        p1.start()
        my_logger.info("starting the acc2tax parsing ...")
    else:
        fp_acc2tax = tempfile.NamedTemporaryFile(delete=False)
        temp_file_list.append(fp_acc2tax)
        my_logger.info("downloading the acc2tax id mappings ...")
        p1 = Process(target=download_via_http_link, args=(taxonomy_acc2taxId,), kwargs={'out_fn': fp_acc2tax.name})
        p1.start(), p1.join()

        p1 = Process(target=read_in_acc_to_taxid, args=(fp_acc2tax.name,), kwargs={"filter_set": sub_set,
                                                                                   "queue": my_queue})
        p1.start()
        my_logger.info("starting the acc2tax parsing ...")
    return p1


def handle_nucleotide_fasta(args, my_logger, my_queue, sub_set, temp_file_list):
    # nucs
    if args.nuc:
        p2 = Process(target=filter_nuc_sequences, args=(args.nuc, sub_set), kwargs={"queue": my_queue})
        p2.start()
        my_logger.info("starting to parse the nuc sequences ...")
    else:
        fp_nucs = tempfile.NamedTemporaryFile(delete=False)
        temp_file_list.append(fp_nucs)
        p2 = Process(target=download_via_http_link, args=(genbank_and_refseq_nucleotide_sequences,),
                     kwargs={'out_fn': fp_nucs.name})
        p2.start(), p2.join()

        p2 = Process(target=filter_nuc_sequences, args=(fp_nucs.name, sub_set), kwargs={"queue": my_queue})
        p2.start()
    return p2


def create_arg_parser(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('out_fn', help='output file name for the FASTA file')
    parser.add_argument("--meta", help='file with ncbi meta data')
    parser.add_argument("--nuc", help='file with ncbi sequence records as FASTA')
    parser.add_argument("--acc2tax", help='file with accession2taxid mapping')
    parser.add_argument("--out_fn_meta", help="output the filtered meta data to this file")
    args = parser.parse_args(argv)
    return args


def main(argv):
    args = create_arg_parser(argv)
    my_logger = create_stdout_logger()

    # meta data
    meta_df = parse_meta_data(args, my_logger)
    sub_set = set(meta_df["#Accession"])

    # tax id and nucs
    my_queue = Queue()
    temp_file_list = list()  # needed as only after the spawned processes are done i can delete these files

    p1 = filter_tax_id_mapping(args, my_logger, my_queue, sub_set, temp_file_list)
    p2 = handle_nucleotide_fasta(args, my_logger, my_queue, sub_set, temp_file_list)

    qe1, qe2 = my_queue.get(), my_queue.get()
    p1.join(), p2.join()
    [(fp.close(), os.remove(fp.name)) for fp in temp_file_list]

    my_logger.info('done with parsing both, writing out the result next')
    # figure out who came out the queue first
    if qe1[0] == 'seq_dic':
        seq_dict, tax_dic = qe1[1], qe2[1]
    elif qe1[0] == 'tax_id':
        seq_dict, tax_dic = qe2[1], qe1[1]

    with open(args.out_fn, 'w') as handle:
        SeqIO.write(my_generator(seq_dict, tax_dic, meta_df), handle, "fasta")
    my_logger.info('done with everything')


if __name__ == "__main__":
    main(sys.argv[1:])
